/// <reference types="node" />
import { EventEmitter } from "events";
import TaskScheduler from "./task_scheduler";
import { Logger } from "../utils/logger";
import { TaskStatus } from "../types";
import FileConcentrator from "./file_concentrator";
export interface DownloaderOptions {
    /** 并发数量 */
    threads: number;
    /** 最大重试次数 */
    retries: number;
    /** 超时阈值 */
    timeout: number;
    /** Custom HTTP Headers */
    headers: string;
    /** 输出目录 */
    output: string;
    /** 是否以数字增序重命名文件 */
    ascending: boolean;
    /** 是否二进制连接下载文件 */
    concat: boolean;
    /** 是否启用调试输出 */
    verbose?: boolean;
    /** 自定义 Logger */
    logger?: Logger;
}
export interface UrlTask {
    /** URL */
    url: string;
}
/**
 * 下载任务结构
 */
export interface DownloadTask extends UrlTask {
    /** 顺序编号 */
    index: number;
    /** 输出文件名 */
    filename: string;
    /** 自定义 Headers */
    headers?: Record<string, string>;
}
export declare class JSONParseError extends Error {
}
export declare class LoadRemoteFileError extends Error {
}
declare class Downloader extends EventEmitter {
    threads: number;
    retries: number;
    timeout: number;
    headers: Record<string, unknown>;
    output: string;
    ascending: boolean;
    concat: boolean;
    verbose: boolean;
    logger: Logger;
    /** 当前运行并发数量 */
    nowRunningThreadsCount: number;
    /** 全部任务数 */
    totalCount: number;
    /** 已完成任务数 */
    finishCount: number;
    /** 放弃任务数 */
    dropCount: number;
    startTime: Date;
    isEnd: boolean;
    /**
     * 所有需要下载的任务
     * 开始后不修改
     */
    tasks: DownloadTask[];
    taskStatusRecord: TaskStatus[];
    scheduler: TaskScheduler<DownloadTask>;
    fileConcentrator: FileConcentrator;
    constructor({ threads, retries, headers, output, ascending, concat, timeout, verbose, logger, }: Partial<DownloaderOptions>);
    onChunkNaming(task: Omit<DownloadTask, "filename">): string;
    /**
     * 从文件添加下载文件 URL
     * @param path 文件路径
     */
    loadUrlsFromFile(path: string): Promise<void>;
    /**
     * 从表达式添加任务
     * @param expression 表达式
     */
    loadUrlsFromExpression(expression: string): void;
    /**
     * 从URL数组添加任务
     * @param urls URL数组
     */
    loadUrlsFromArray(urls: string[]): void;
    /**
     * 从 JSON 文件添加任务
     * @param path
     */
    loadUrlsFromJSON(path: string): void;
    setupTaskScheduler(): void;
    addTasks(tasks: UrlTask[]): void;
    /**
     * 开始下载
     */
    start(): void;
    handleTask(task: DownloadTask): Promise<void>;
    beforeFinish(): Promise<void>;
    getETA(): string;
}
export default Downloader;
